% Three-dimensional, incomplete data FDE test

%%%%%%%%%%% Setup part begin %%%%%%%%%%


alpha = 0.5;    % Temporal derivative order
beta1 = 1.5; beta2 = 1.5; beta3 = 1.5;  % Space derivative orders

L = 1;      % length of spatial interval
d = 5;      % grid level
nx = 2^d+2; % Number of spatial steps of discretization
ny = 2^d+2;
nz = 2^d+2;

nt = 2^(2*d);   % Number of steps in time


tol = 1e-6;     % tensor approximation for the solution

eps = 1e-14;    % tensor approximation for matrices.
                % It should be as small as possible, but
                % it cannot be the machine eps due to numerical SVD noise

gamma=1e-6;     % Regularization parameter 
                % !!!! we will divide over it, do not set zero !!!

% Rectangular observation domain, defined by left and right grid points,
% i.e. [ax_obs,bx_obs] x [ay_obs,by_obs] x [az_obs,bz_obs]
% The values should be given, assuming the boundary points excluded!

ax_obs = 2^(d-1)-2^(d-3);   % left bound for observation domain
bx_obs = 2^(d-1)+2^(d-3)-1; % right bound for observation domain
ay_obs = 2^(d-1)-2^(d-3);   % lower bound for obs domain
by_obs = 2^(d-1)+2^(d-3)-1; % upper bound for obs domain
az_obs = 2^(d-1)-2^(d-3);   % lower bound for obs domain
bz_obs = 2^(d-1)+2^(d-3)-1; % upper bound for obs domain

% Rectangular control domain, defined by left and right grid points,
% i.e. [ax_con,bx_con] x [ay_con,by_con] x [az_con,bz_con]
% The values should be given, assuming the boundary points excluded!

ax_con = 2^(d-1)-2^(d-3);   % left bound for observation domain
bx_con = 2^(d-1)+2^(d-3)-1; % right bound for observation domain
ay_con = 2^(d-1)-2^(d-3);   % lower bound for obs domain
by_con = 2^(d-1)+2^(d-3)-1; % upper bound for obs domain
az_con = 2^(d-1)-2^(d-3);   % lower bound for obs domain
bz_con = 2^(d-1)+2^(d-3)-1; % upper bound for obs domain


%%%%%%%%%%% Setup part end %%%%%%%%%%

% Auto-initialise TT-Toolbox
if (exist('tt_tensor', 'file')==0)    
    % Check whether the TT-Toolbox directory is already here
    if (exist('TT-Toolbox-small', 'dir')~=7)
        fprintf('Downloading and unpacking TT-Toolbox...\n');
        [f,stat] = urlwrite('https://codeload.github.com/oseledets/TT-Toolbox/zip/small', 'TT-Toolbox-small.zip');
        if (stat==0)
            error('Sorry, automatic downloading failed. You need to download http://github.com/oseledets/TT-Toolbox/archive/small.zip and unzip it manually.');
        end;
        unzip('TT-Toolbox-small.zip');
        delete('TT-Toolbox-small.zip');
    end;
    fprintf('Adding TT-Toolbox to the path...\n');
    cd TT-Toolbox-small
    setup
    cd ..
    fprintf('TT-Toolbox is ready\n');
end;

hx = L/(nx-1);  % spatial mesh steps
hy = L/(ny-1);
hz = L/(nz-1);

tau = hx*hy;    % time step

% Quantized sizes
mx = factor(nx-2); % space-x size
my = factor(ny-2); % space-y size
mz = factor(nz-2); % space-z size
mt = factor(nt);   % temporal size
mx_c = factor(bx_con-ax_con+1); % control dimension x
my_c = factor(by_con-ay_con+1); % control dimension y
mz_c = factor(bz_con-az_con+1); % control dimension z



% Coefficient vectors by Grunwald-Letnikov formulae
v_alpha = cumprod([1, 1 - ((alpha+1) ./ (1:nt))]);
v_beta1 =  cumprod([1, 1 - ((beta1+1) ./ (1:nx))]);
v_beta2 = cumprod([1, 1 - ((beta2+1) ./ (1:ny))]);
v_beta3 = cumprod([1, 1 - ((beta3+1) ./ (1:nz))]);

% Temporal Toeplitz matrix in the QTT format
% Compress v_alpha i.e. the first column
tt_v_alpha = tt_tensor(v_alpha(1:end-1), eps, mt);
% The tt_qtoepl routine puts elements (n+1)--(2n) to the column. 
% Shift v_alpha accordingly
tt_v_alpha = tkron(tt_v_alpha, tt_tensor([0;1]));
% Create the Toeplitz matrix
tt_Talpha = (tau^(-alpha))*tt_qtoepl(tt_v_alpha, log2(nt));
tt_Talpha = round(tt_Talpha, eps);

% Spatial matrices in the QTT format. 
% Boundary elements are already excluded.
tt_v_beta1 = tt_tensor(v_beta1(2:end-2), eps, mx);
% Shift to the range (n+1)--(2n)
tt_v_beta1 = tkron(tt_v_beta1, tt_tensor([0;1]));
% Add one element above the diagonal
tt_v_beta1 = tt_v_beta1 + v_beta1(1)*tt_unit(2,d+1,2^d);
tt_v_beta1 = round(tt_v_beta1, eps);
% Create the Toeplitz matrix
tt_C_beta1 = (hx^(-beta1))*tt_qtoepl(tt_v_beta1, d);
tt_L_beta1 = (tt_C_beta1+tt_C_beta1')*0.5;
tt_L_beta1 = round(tt_L_beta1, eps);

% Repeat the same steps for other coordinates
tt_v_beta2 = tt_tensor(v_beta2(2:end-2), eps, my);
tt_v_beta2 = tkron(tt_v_beta2, tt_tensor([0;1]));
tt_v_beta2 = tt_v_beta2 + v_beta2(1)*tt_unit(2,d+1,2^d);
tt_v_beta2 = round(tt_v_beta2, eps);
tt_C_beta2 = (hy^(-beta2))*tt_qtoepl(tt_v_beta2, d);
tt_L_beta2 = (tt_C_beta2+tt_C_beta2')*0.5;
tt_L_beta2 = round(tt_L_beta2, eps);

tt_v_beta3 = tt_tensor(v_beta3(2:end-2), eps, mz);
tt_v_beta3 = tkron(tt_v_beta3, tt_tensor([0;1]));
tt_v_beta3 = tt_v_beta3 + v_beta3(1)*tt_unit(2,d+1,2^d);
tt_v_beta3 = round(tt_v_beta3, eps);
tt_C_beta3 = (hz^(-beta3))*tt_qtoepl(tt_v_beta3, d);
tt_L_beta3 = (tt_C_beta3+tt_C_beta3')*0.5;
tt_L_beta3 = round(tt_L_beta3, eps);


% A tensor-structured version of meshgrid
tt_X = (tt_x(2,d)+1)*hx;
tt_X = mtkron(tt_X, tt_ones(2,d), tt_ones(2,d));
tt_Y = (tt_x(2,d)+1)*hy;
tt_Y = mtkron(tt_ones(2,d), tt_Y, tt_ones(2,d));
tt_Z = (tt_x(2,d)+1)*hz;
tt_Z = mtkron(tt_ones(2,d), tt_ones(2,d), tt_Z);

% The forcing vector is zero
tt_F = tt_zeros(2, d*3);

% The target vector is constructed as a function of X,Y,Z via the cross
% interpolation.
tt_YD = amen_cross({tt_X,tt_Y,tt_Z}, @(X)exp(-64*((X(:,1)-0.5).^2+(X(:,2)-0.5).^2+(X(:,3)-0.5).^2)), ...
    eps, 'y0', mtkron(tt_unit(2,d,2^d/2),tt_unit(2,d,2^d/2),tt_unit(2,d,2^d/2)));


% Create an observation operator
% Prepare the vectors of ones between ax and bx and zeros otherwise
c1 = zeros(nx-2, 1);
c1(ax_obs:bx_obs) = 1;
% And compress them to tt_tensors.
c1 = tt_tensor(c1, eps, mx);

c2 = zeros(ny-2, 1);
c2(ay_obs:by_obs) = 1;
c2 = tt_tensor(c2, eps, my);

c3 = zeros(nz-2, 1);
c3(az_obs:bz_obs) = 1;
c3 = tt_tensor(c3, eps, mz);

% Put them onto a diagonal
tt_C1 = diag(c1);
tt_C2 = diag(c2);
tt_C3 = diag(c3);


% Create a control operator
% These matrices are rectangular, prepare them as such, putting an identity
% matrix to the block corresponding to [ax,bx], etc.
p1=sparse(nx-2,bx_con-ax_con+1);
p1(ax_con:bx_con,:) = speye(bx_con-ax_con+1);
p2=sparse(ny-2,by_con-ay_con+1);
p2(ay_con:by_con,:) = speye(by_con-ay_con+1);
p3=sparse(nz-2,bz_con-az_con+1);
p3(az_con:bz_con,:) = speye(bz_con-az_con+1);
% Now they should be unfortunately compressed from the full format, since the
% system and control sizes can have different factors
tt_p1 = tt_matrix(full(p1), eps, mx, mx_c);
tt_p2 = tt_matrix(full(p2), eps, my, my_c);
tt_p3 = tt_matrix(full(p3), eps, mz, mz_c);


% Spatial FDE matrix = L1 x I x I + I x L2 x I + I x I x L3
Lspace = mtkron(tt_L_beta1, tt_eye(my), tt_eye(mz))+...
    mtkron(tt_eye(mx), tt_L_beta2, tt_eye(mz))+...
    mtkron(tt_eye(mx), tt_eye(my), tt_L_beta3);

% Temporal FDE matrix = (IxIxI) x T
Ltime = tkron(tt_eye([mx,my,mz]), tt_Talpha);
% Spatial matrix extended by I in the time variable
A = tkron(Lspace,tt_eye(mt));
% Finally, the whole system matrix is as follows
A=(hx*hy*hz*tau)*round(Ltime-A,eps);


% Observation operator -- the kron of Ci and the identity in time.
M1 = mtkron(tt_C1, tt_C2, tt_C3, tt_eye(mt));
M1 = (hx*hy*hz*tau)*M1;

% Rectangular control--system "interaction" operator
M3 = mtkron(tt_p1, tt_p2, tt_p3, tt_eye(mt));
M3 = (hx*hy*hz*tau)*M3;

% Control mass matrix
M2 = tt_eye([mx_c, my_c, mz_c, mt])*(hx*hy*hz*tau*gamma);

% Right hand side, first block
pre_f1 = tkron(tt_YD, tt_ones(mt));
f1 = M1*pre_f1;
% The second block is always zero

% Third block
pre_f3 = tkron(tt_F, tt_ones(mt));
f3 = hx*hy*hz*tau*pre_f3;


% TT Solver -- SC2
fprintf('\n======== Second Schur complement\n');
tic;
fprintf('f3t started...\n');
f3t = amen_solve2(A, f3, tol, 'trunc_norm', 'fro');
f1t = f1 - M1*f3t;  % Shorter than in the paper, use that f2 = 0, g2 = 0
f1t = round(f1t, tol);
% Now compute the heavy part of the Schur Complement matrix
% First, M2^{-1}M3^T
M2h = M2;
M2h = tt_reshape(M2h, M3.m*[1,1]);
M2h = kron2(M2h, tt_eye(M3.n));
fprintf('inv(M2)*M3^T started...\n');
iMN = amen_solve2(M2h, tt_tensor(M3'), tol, 'trunc_norm', 'fro', 'x0', tt_tensor(M3'));
iMN = tt_matrix(iMN, M3.m, M3.n);
% Compute M3*inv(M2)*M3^T
NiMN = M3*iMN;
NiMN = round(NiMN, tol);
% And apply A^{-1} to this
Ah = kron2(A, tt_eye(M3.n));
fprintf('inv(A)*[M3*inv(M2)*M3^T] started...\n');
SC = amen_solve2(Ah, tt_tensor(NiMN), tol, 'trunc_norm', 'fro', 'kickrank', 10, 'x0', tt_tensor(NiMN));
SC = tt_matrix(SC, M3.n, M3.n);
% Now, the Schur Complement matrix is almost ready. It reads
% M1*A^{-1}*M3*M2^{-1}*M3^T+A^T
SC = M1*SC;
SC = round(SC, tol);
SC = SC + A';
fprintf('Schur complement is ready\n');
ttimes(1,2)=toc;

% One more heavy step - solve for p
tic;
fprintf('p started...\n');
p_sc2 = amen_solve2(SC, f1t, tol, 'trunc_norm', 'fro', 'x0', f1t, 'local_restart', 100, 'nswp', 20);
ttimes(2,2)=toc;
tic;
% f2 = 0
u_sc2 = -M3'*p_sc2;
M2r = M2;
M2r = tt_reshape(M2r, M3.m*[1,1]);
fprintf('u started...\n');
u_sc2 = amen_solve2(M2r, u_sc2, tol, 'trunc_norm', 'fro', 'x0', u_sc2);
fprintf('y started...\n');
y_sc2 = amen_solve2(A, f3 - M3*u_sc2, tol, 'trunc_norm', 'fro', 'nswp', 30);
ttimes(3,2)=toc;


% Measure each row in the residual
resid_row1 = norm(M1*y_sc2+A'*p_sc2 - f1)/norm(f1)
resid_row2 = norm(M2r*u_sc2+M3'*p_sc2)/norm(M3'*p_sc2)
resid_row3 = norm(A*y_sc2 + M3*u_sc2-f3)/max(norm(M3*u_sc2), norm(f3))

% A discrepancy with the target
obs_error = norm(f1-M1*y_sc2)/norm(f1)

fprintf('\n======== Second SC completed, timings: [%g,%g,%g]\n', ttimes(1,2),ttimes(2,2),ttimes(3,2));


% For small grids, we can visualise this
if (d<=5)
    % Create the grid
    [X2,Y2] = meshgrid(hx:hx:L-hx,hy:hy:L-hy);
    % A matrix for extracting the last time snapshot
    P = mtkron(tt_eye(mx), tt_eye(my), tt_eye(mz), tt_matrix(tt_unit(mt, numel(mt), nt), 1, mt));
    
    % State in the full format
    y_full = P*(M1/(hx*hy*hz*tau))*y_sc2;
    y_full = full(y_full, [nx-2,ny-2,nz-2]);
    figure(1);
    mesh(X2,Y2,y_full(:,:,nz/2-1));
    title('y');
    
    % Target in the full format
    yd_full = P*f1/(hx*hy*hz*tau);
    yd_full = full(yd_full, [nx-2,ny-2,nz-2]);
    figure(2)
    mesh(X2,Y2,yd_full(:,:,nz/2-1));
    title('yd');
    
    % Control
    u_full = P*(M3/(hx*hy*hz*tau))*u_sc2;
    u_full = full(u_full, [nx-2,ny-2,nz-2]);
    figure(3)
    mesh(X2,Y2,u_full(:,:,nz/2-1));
    title('u');
    
    % Lagrange multipliers
    p_full = P*p_sc2;
    p_full = full(p_full, [nx-2,ny-2,nz-2]);
    figure(4)
    mesh(X2,Y2,p_full(:,:,nz/2-1));
    title('p');
end;

